Unit TextViewer;

INTERFACE

USES
   Classes, SysUtils, Controls, Graphics, StdCtrls, Forms, Messages,
   Contnrs;

TYPE
//********************************************************************
   TStyleInfo = Class(TObject)
      fFontIndex: Byte;       // Index der Font im Font-Array
      fPrintPosY: Integer;    // Printposition der jeweiligen Zeile
      fMaxTextHeight: Byte;     // Höhe der größten Schrift in dieser
   End;                       //                    Zeile
//********************************************************************
   TTextViewer = class(TObject)
   Private
      fForm: TForm;
      fVScrollBar: TScrollBar;      // vertikale ScrollBar

      fTextBuffer: TStringList;
      fStyleInfoList: TObjectList;  // Eine Liste für die Style-Infos

      fViewerFont: Array Of TFont;  // Font-Array
      fCanvasFont: TFont;           // Sicherung der Canvas-Font
      fFontCount: Byte;

      fFirstLine: Integer;
      fLastLine: Integer;

      Procedure PageUp;
      Procedure PageDown;

      Procedure SetupTextStyle;
       Function GetFirstLine(APixelPos: Integer): Integer;
      Procedure InitLines;
   Public
      Constructor Create(AForm: TForm;
                      AVertScrollBar: TScrollBar); Reintroduce;
      Destructor Destroy; Override;
      Procedure Load(AFileName: AnsiString);
      Procedure PrintPage;
      Procedure Scroll(Sender: TObject;
          AScrollCode: TScrollCode; AScrollPos: Integer);
      Procedure SetupScrollBars;
  End;

IMPLEMENTATION

//********************************************************************
Constructor TTextViewer.Create(AForm: TForm; AVertScrollBar: TScrollBar);
Begin
  Inherited Create;

  fForm:= AForm;
  fCanvasFont:= fForm.Canvas.Font;

  fVScrollBar:= AVertScrollBar;
  fTextBuffer:=TStringList.Create;
  fStyleInfoList:= TObjectList.Create;

  // Zwei verschiedene Fonts erzeugen
  fFontCount:= 3;
  SetLength(fViewerFont,fFontCount);

  fViewerFont[0]:= TFont.Create;
  With fViewerFont[0] Do Begin
     Color:= clRed;
     Size:= 24;
  End;
  fViewerFont[1]:= TFont.Create;
  With fViewerFont[1] Do Begin
     Color:= clBlue;
     Size:= 12;
  End;

  fViewerFont[2]:= TFont.Create;
  With fViewerFont[2] Do Begin
     Color:= clLime;
     Size:= 25;
  End;

End;
//********************************************************************
Destructor TTextViewer.Destroy;
Var I: Byte;
Begin
   fForm.Canvas.Font:= fCanvasFont;

   fStyleInfoList.Free;
   fTextBuffer.Free;
   For I:= 0 To fFontCount-1 Do
      If fViewerFont[I] <> NIL
         Then fViewerFont[I].Free;

   Inherited Destroy;
End;
//********************************************************************
Procedure TTextViewer.Load(AFileName: AnsiString);
Begin
   fTextBuffer.LoadFromFile(AFileName);

   // Textstyles erzeugen
   SetupTextStyle;

   // Scrollbar initialisieren
   SetupScrollBars;
   fFirstLine:= 0;
End;
//********************************************************************
Procedure TTextViewer.PrintPage;
Var
   Styleinfo: TStyleInfo;
   CurLine,DiffPos,PrintPosX,PrintPosY,I: Integer;
Begin
   With fForm,fTextBuffer Do Begin
      StyleInfo:= fStyleInfoList.List[fFirstLine];
      DiffPos:= StyleInfo.fPrintPosY;
      PrintPosX:= 0;
      PrintPosY:= 0;
      I:= 0;
      While (PrintPosY < fForm.ClientHeight) And
            (fFirstLine+I < fTextBuffer.Count) Do
         Begin
            CurLine:= fFirstLine+I;

            StyleInfo:= fStyleInfoList.List[CurLine];
            PrintPosY:= StyleInfo.fPrintPosY - DiffPos + 1;
            Canvas.Font:= fViewerFont[StyleInfo.fFontIndex];

            Canvas.TextOut(PrintPosX,PrintPosY,Strings[CurLine]);

            Inc(I);
         End;
   End;
End;
//********************************************************************
Procedure TTextViewer.SetupScrollBars;
Var StyleInfo: TStyleInfo;
Begin
   With fStyleInfoList Do StyleInfo:= List[Count-1];

   With fVScrollBar Do Begin
      Min:= 0;
      Max:= StyleInfo.fPrintPosY + StyleInfo.fMaxTextHeight;
      LargeChange:= fForm.ClientHeight;
      SmallChange:= 20; // Dummy-Wert
   End;

   InitLines;
End;
//********************************************************************
Procedure TTextViewer.Scroll(Sender: TObject;
          AScrollCode: TScrollCode; AScrollPos: Integer);
Begin
   With fTextBuffer Do Case AScrollCode of

        scTop: fFirstLine:= 0;
     scBottom: fFirstLine:= fTextBuffer.Count-1;
     scLineUp: If fFirstLine > 0
                   Then Dec(fFirstLine);
   scLineDown: If fFirstLine < fLastLine
                   Then Inc(fFirstLine);
     scPageUp: Begin
                  PageUp;
                  If fFirstLine < 0
                     Then fFirstLine:= 0;
               End;
   scPageDown: Begin
                  PageDown;
                  If fFirstLine > fLastLine
                     Then fFirstLine:= fLastLine;
               End;
   scPosition: Begin
                  fFirstLine:= GetFirstLine(AScrollPos);
                  If fFirstLine > fLastLine
                     Then fFirstLine:= fLastLine;
                  If fFirstLine = -1
                     Then fFirstLine:= 0;
               End;

   else exit;

   End; // Of Case

   // Form neu zeichnen ****************
   fForm.Refresh;
End;
//********************************************************************
Procedure TTextViewer.PageUp;
Var Styleinfo: TStyleInfo; DiffPos: Integer;
Begin
   DiffPos:= 0;

   While (DiffPos <= fForm.ClientHeight) And (fFirstLine >= 0) Do
      Begin
         StyleInfo:= fStyleInfoList.List[fFirstLine];
         Inc(DiffPos,StyleInfo.fMaxTextHeight);
         Dec(fFirstLine);
      End;
   Inc(fFirstLine);
End;
//********************************************************************
Procedure TTextViewer.PageDown;
Var Styleinfo: TStyleInfo; DiffPos: Integer;
Begin
   DiffPos:= 0;

   While (DiffPos <= fForm.ClientHeight) And
         (fFirstLine <= fTextBuffer.Count-1) Do
      Begin
         StyleInfo:= fStyleInfoList.List[fFirstLine];
         Inc(DiffPos,StyleInfo.fMaxTextHeight);
         Inc(fFirstLine);
      End;
   Dec(fFirstLine);
End;
//********************************************************************
// Jede Zeile erhält eine zufällige Font
Procedure TTextViewer.SetupTextStyle;
Const INTERIM = 8;
Var StyleInfo: TStyleInfo; RDNumber: Single; CurPos,I: Integer;
Begin
  CurPos:= 0;

  For I:= 0 To fTextBuffer.Count-1 Do Begin

     StyleInfo:= TStyleInfo.Create;
//     RDNumber:= ;  // Zufallsgenerator

     With StyleInfo Do Begin
        fFontIndex:=Random(3);
        fPrintPosY:= CurPos;
        fMaxTextHeight:= fViewerFont[fFontIndex].Size+INTERIM;
     End;

     Inc(CurPos,StyleInfo.fMaxTextHeight);
     fStyleInfoList.Add(StyleInfo);
  End;
End;
//********************************************************************
// Binärerer Suchalgorithmus
Function TTextViewer.GetFirstLine(APixelPos: Integer): Integer;
Type
   TSearchResult = (srEqual,srBigger,srSmaller);
Var
   StyleInfo: TStyleInfo;
   First: Integer;
   Last: Integer;
   Pivot: Integer;
   SearchResult: TSearchResult;
   Found: Boolean;

//--------------------------------------------------------------
Function CheckFound(APixelPos,ALinePos: Integer): TSearchResult;
Var Diff: Integer;
Begin
   Result:= srBigger;

   Diff:= APixelPos-ALinePos;
   If (Diff >= 0) And (Diff <= StyleInfo.fMaxTextHeight)
      Then Result:= srEqual
      Else if Diff < 0
         Then Result:= srSmaller;
End;
//--------------------------------------------------------------
Begin
   First := 0;
   Last  := fTextBuffer.Count-1;
   Found := FALSE;
   Result:= -1;

   While (First <= Last) And (Not Found) Do
   Begin
     Pivot := (First + Last) Div 2;
     With fStyleInfoList Do StyleInfo:= List[Pivot];
     SearchResult:= CheckFound(APixelPos,StyleInfo.fPrintPosY);
     If SearchResult = srEqual
        Then Begin
          Found := True;
          Result:= Pivot;
        End
        Else If SearchResult = srBigger
           Then First:= Pivot + 1
           Else Last := Pivot - 1;
   End;
End;
//********************************************************************
// Letzte Zeile bestimmen
Procedure TTextViewer.InitLines;
Var SaveFirstLine: Integer;
Begin
   SaveFirstLine:= fFirstLine;
   fFirstLine:= fTextBuffer.Count-1;
   Pageup;
   fLastLine:= fFirstLine+1;
   fFirstLine:= SaveFirstLine;
End;
//********************************************************************
End.
